# encoding: utf-8
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#
"""
Creates a loan request using information provided by a web formular.
"""

from __future__ import absolute_import, annotations

import logging
import sys
import traceback
from datetime import datetime

import caosdb as db
from caosadvancedtools.serverside.helper import (DataModelError, get_data,
                                                 init_data_model,get_timestamp,
                                                 parse_arguments, print_error,
                                                 print_info, print_warning,
                                                 send_mail)
from caosadvancedtools.serverside.logging import configure_server_side_logging
from linkahead.exceptions import EmptyUniqueQueryError
from validate_email import validate_email
from .conf import *

LOGGER_NAME = "box_loan"
LOGGER = logging.getLogger(LOGGER_NAME)

# Form names
F_BOX = "box"
F_FIRST_NAME = "first_name"
F_LAST_NAME = "last_name"
F_EMAIL = "email"
F_EXPECTED_RETURN_DATE = "expected_return_date"
F_EXHAUST_CONTENTS = "exhaust_contents"
F_COMMENT = "comment"
F_DESTINATION = "destination"
F_LOAN = "loan"
F_CURRENT_LOCATION = "current_location"

S_RETURNED = "returned"
S_RETURN_ACCEPTED = "return_accepted"
S_RETURN_REQUESTED = "return_requested"
S_LENT = "lent"
S_LOAN_ACCEPTED = "loan_accepted"
S_LOAN_REQUESTED = "loan_requested"


_EMTPY_DATA = {
    F_BOX: None,
    F_FIRST_NAME: None,
    F_LAST_NAME: None,
    F_EMAIL: None,
    F_EXPECTED_RETURN_DATE: None,
    F_EXHAUST_CONTENTS: None,
    F_COMMENT: None,
    F_DESTINATION: None,
    F_LOAN: None,
    F_CURRENT_LOCATION: None
}


RECORD_TYPES = [
    BOX,
    PERSON,
    LOAN,
]


PROPERTIES = [
    FIRST_NAME,
    LAST_NAME,
    EMAIL,
    LOCATION,
    DESTINATION,
    RETURNLOCATION,
    COMMENT,
    EXHAUST_CONTENTS,
    BORROWER,
    CONTENT,
    LOAN_REQUESTED,
    EXPECTED_RETURN,
    LOAN_ACCEPTED,
    LENT,
    RETURN_REQUESTED,
    RETURN_ACCEPTED,
    RETURNED,
    BOX_NUMBER,
]


def _get_mail_from():
    config = db.get_config()
    result = "admin@indiscale.com"

    if "Misc" in config and "entity_loan.curator_mail_from" in config["Misc"]:
        result = config["Misc"]["entity_loan.curator_mail_from"]

    return result


def _get_mail_curator():
    config = db.get_config()
    result = "admin@indiscale.com"

    if "Misc" in config and "entity_loan.curator_mail_to" in config["Misc"]:
        result = config["Misc"]["entity_loan.curator_mail_to"]

    return result


EMAIL_FROM = _get_mail_from()
EMAIL_CURATOR = _get_mail_curator()


class MissingParentError(RuntimeError):

    def __init__(self, entity, expected_parent):
        super().__init__(
            "This entity should have a {par} parent:\n{ent}".format(
                par=expected_parent.name,
                ent=str(entity)))


class StateError(RuntimeError):

    def __init__(self, actual, expected):
        one = "" if len(expected) == 1 else "one of "
        _expected = expected[0] if len(expected) == 1 else ",".join(
            expected[:-1]) + ", or" + expected[-1]
        super().__init__(
            "This transition of the loan state is not possible. The allowed target "
            "state is {one}'{expected}' "
            "while the actual target state is '{actual}'.".format(one=one,
                                                           actual=actual,
                                                           expected=_expected))


class EmailPatternError(RuntimeError):

    def __init__(self, email):
        super().__init__("The email address '{}' is not valid.".format(email))


class DataError(RuntimeError):

    def __init__(self, msg):
        super().__init__("The request most likely contained a data error."
                         " {}Please check your input.".format(msg if msg is not None else ""))


def assert_key_in_data(data, key):
    if key not in data:
        raise DataError("Missing information. "
                        "The data is expected to contain a `{}` "
                        "property.".format(key))


def assert_is_date(string, msg=None):
    try:
        return datetime.fromisoformat(string)
    except (TypeError, ValueError):
        pass  # and raise another error

    _msg = msg if msg is not None else ("Should be a ISO 8601 date "
                                        "string. "
                                        "Was '{}'. ".format(string))
    raise DataError(_msg)


def assert_date_in_future(actual, msg=None):
    date = assert_is_date(actual)

    if date <= datetime.now(tz=date.tzinfo):
        _msg = msg if msg is not None else ("Date is to be in the future. "
                                            "Was '{}'. ".format(actual))
        raise DataError(_msg)


def get_record_by_id(parent, entity_id):
    """ Retrieve a record with a particular parent by id. """

    return db.execute_query("FIND RECORD {} WITH ID = {}".format(parent,
                                                                 entity_id),
                            unique=True)


def get_box(box):
    """ Retrieve a box record by id. """

    return get_record_by_id(BOX.name, box)


def get_loan(loan_id):
    """ Retrieve a loan record by id. """

    return get_record_by_id(LOAN.name, loan_id)


def update_loan_state(loan, new_state):
    old_state = get_loan_state(loan)
    if old_state==S_LOAN_REQUESTED:
        if new_state==S_LOAN_ACCEPTED:
            loan.add_property(LOAN_ACCEPTED, get_timestamp())
        else:
            raise StateError(new_state, [S_LOAN_ACCEPTED])
    elif old_state==S_LOAN_ACCEPTED:
        if new_state==S_LENT:
            loan.add_property(LENT, get_timestamp())
        else:
            raise StateError(new_state, [S_LENT])
    elif old_state==S_LENT:
        if new_state==S_RETURN_REQUESTED:
            loan.add_property(RETURN_REQUESTED, get_timestamp())
        else:
            raise StateError(new_state, [S_RETURN_REQUESTED])
    elif old_state==S_RETURN_REQUESTED:
        if new_state==S_RETURN_ACCEPTED:
            loan.add_property(RETURN_ACCEPTED, get_timestamp())
        elif new_state==S_LENT:
            loan.remove_property(RETURN_REQUESTED)
        else:
            raise StateError(new_state, [S_RETURN_ACCEPTED, S_LENT])
    elif old_state==S_RETURN_ACCEPTED:
        if new_state==S_RETURNED:
            loan.add_property(RETURNED, get_timestamp())
        else:
            raise StateError(new_state, [S_RETURNED])
    else:
        raise RuntimeError(f"Old state {old_state} is unknown.")


def get_loan_state(loan):
    if loan.get_parent(LOAN) is None:
        raise MissingParentError(loan, LOAN)

    if loan.get_property(RETURNED.name) is not None:
        return S_RETURNED

    if loan.get_property(RETURN_ACCEPTED.name) is not None:
        return S_RETURN_ACCEPTED

    if loan.get_property(RETURN_REQUESTED.name) is not None:
        return S_RETURN_REQUESTED

    if loan.get_property(LENT.name) is not None:
        return S_LENT

    if loan.get_property(LOAN_ACCEPTED.name) is not None:
        return S_LOAN_ACCEPTED

    if loan.get_property(LOAN_REQUESTED.name) is not None:
        return S_LOAN_REQUESTED


def assert_loan_state(loan, state):
    actual = get_loan_state(loan)
    expected = state if isinstance(state, list) else [state]

    if actual not in expected:
        raise StateError(actual, expected)


def get_borrower_names(loan):
    """Retrieve the borrower's first name and last name.

    Intented for generating a nice and informative success message.

    If the loan has no borrower ["an", "unknown borrower"] is returned.

    Parameters
    ----------
    loan : db.Record
        A `loan` Record which should have a `borrower` Property.
    """
    borrower_property = loan.get_property(BORROWER.name)

    if borrower_property is None:
        # The loan had no borrower. This should not happen, but we don't want
        # it to break the functionality.
        # TODO issue warning
        fn = "an"
        ln = "unknown borrower"
    else:
        borrower = db.Record(id=borrower_property.value).retrieve()
        fn = borrower.get_property(FIRST_NAME.name).value
        ln = borrower.get_property(LAST_NAME.name).value

    return fn, ln


def assert_email_pattern(email):
    """Raise an EmailPatternError if the email does't match.

    Parameters
    ----------
    email: str

    Raises
    ------
    EmailPatternError
    """

    if email is None or not validate_email(
            email_address=email,
            check_format=True,
            check_blacklist=False,
            check_dns=False,
            check_smtp=False):
        raise EmailPatternError(email)


def get_property_value(entity, prop, fallback=None):
    p = entity.get_property(prop)

    if p is None or p.value is None:
        return fallback

    return p.value


def get_requester_string(requester):
    results = [
        get_property_value(requester, FIRST_NAME, fallback=""),
        get_property_value(requester, LAST_NAME, fallback=""),
        get_property_value(requester, EMAIL, fallback="")
    ]

    # wrap email in < and >

    if len(results[2]) > 0:
        results[2] = "<" + results[2] + ">"

    return " ".join(results)


def get_external_server_uri():
    uri = db.get_config().get("sss_helper", "external_uri")

    if uri is not None:
        if not uri.endswith("/"):
            uri = uri + "/"

        return uri
    raise RuntimeError("Could not determine the external server uri")


def send_loan_request_mail(data, borrower, loan):
    try:
        boxid = str(loan.get_property(BOX).value).split("@")[0]
        box = get_box(boxid)
        boxnumber = get_property_value(box, BOX_NUMBER, "UNKNOWN")

        link = "{uri}Entity/{boxid}&{loanid}".format(
            uri=get_external_server_uri(),
            boxid=boxid,
            loanid=loan.id)

        body = """Dear Curator,

    a new loan has been request by {borrower} for box number {boxnumber}.

    Loan request:
        {data}

    View box and loan records:
        {link}
    """.format(boxnumber=boxnumber, borrower=get_requester_string(borrower),
               data=data, link=link)

        send_mail(
            from_addr=EMAIL_FROM,
            to=EMAIL_CURATOR,
            subject="loan request",
            body=body)
        print_info("An email has been sent to the responsible box curator.")
    except BaseException as e:
        print_error("""Sending an email to the responsible box curator failed

for an unknown reason. Please inform the box curator by yourself about your
loan request and that the email sending failed. Thank you very much for your
help.""")
        LOGGER.error(e)


def send_return_request_mail(data, returner, loan):
    try:
        boxid = str(loan.get_property(BOX).value).split("@")[0]
        box = get_box(boxid)
        boxnumber = get_property_value(box, BOX_NUMBER, "UNKNOWN")
        current_location = data[F_CURRENT_LOCATION]

        link = "{uri}Entity/{boxid}&{loanid}".format(
            uri=get_external_server_uri(),
            boxid=boxid,
            loanid=loan.id)

        body = """Dear Curator,

    a new return request by {returner} for box number {boxnumber} is pending.
    The current location of the box according to the return request is
    "{current_location}".

    Return request:
        {data}

    View box and loan records:
        {link}
    """.format(boxnumber=boxnumber, returner=get_requester_string(returner),
               data=data, link=link, current_location=current_location)

        send_mail(
            from_addr=EMAIL_FROM,
            to=EMAIL_CURATOR,
            subject="return request",
            body=body)
        print_info("An email has been sent to the responsible box curator.")
    except BaseException as e:
        print_error("""Sending an email to the responsible box curator failed

for an unknown reason. Please inform the box curator by yourself about this

return request and that the email sending failed. Thank you very much for your
help.""")
        LOGGER.error(e)


def create_person(firstname, lastname, email):
    """ Insert a new person Record."""
    person = db.Record().add_parent(PERSON)
    person.add_property(FIRST_NAME, firstname)
    person.add_property(LAST_NAME, lastname)
    assert_email_pattern(email)
    person.add_property(EMAIL, email)

    return person


def query_person(firstname, lastname, email):
    """ Query a person

    The person is identified by
    1. the email address
    2. or the first name and the last name
    """

    query=(f'FIND RECORD {PERSON.name} '
             f'WITH {EMAIL.name} = "{email}" '
             f'OR ({FIRST_NAME.name} = "{firstname}" AND {LAST_NAME.name} = "{lastname}") '
             )
    return db.execute_query(query, unique=True)




def _update_person(person, firstname, lastname, email):
    """Update the persons names and email address.

    But only if anything changes. That said, if only the case of the names or
    the email address changes, the person entity is left unchanged.
    """

    if set_property(
            person, EMAIL, email, False) or set_property(
            person, FIRST_NAME, firstname, False) or set_property(
            person, LAST_NAME, lastname, False):
        person.update()


def insert_or_update_person(firstname, lastname, email):
    """ Retrieve a person Record or create it. """
    try:
        person = query_person(firstname, lastname, email)

        # TODO this is a side effect which might not be desired in all
        # cases. How should we handle it?
        _update_person(person, firstname, lastname, email)

        return person
    except EmptyUniqueQueryError:
        return create_person(firstname, lastname, email).insert()


def set_property(entity, prop, value, case_sensitive=True):
    """Set a property value.

    If the property already has that value, the value is left untouched.

    Parameters
    ----------
    entity : db.Entity
    prop : str or db.Property
    value : object
    case_sensitive : boolean (optional)
        defaults to 'True'. Only matters if `value` is a string.
    """
    p = entity.get_property(prop)

    if p is None and value is not None:
        entity.add_property(prop, value)

        return True
    elif ((case_sensitive is True and p.value != value)
          or (hasattr(p.value, "lower")
              and hasattr(value, "lower")
              and p.value.lower() != value.lower())):
        p.value = value

        return True

    return False


def check_data_model():
    # TODO init_data_model is a bad name; no initialization is done
    return init_data_model(RECORD_TYPES + PROPERTIES)


def _caller(func, args):
    params = parse_arguments(args)

    if params.auth_token:
        db.configure_connection(password_method="auth_token",
                                auth_token=params.auth_token)
        check_data_model()
    else:
        print_warning("No auth_token passed.")

    data = get_data(params.filename, default=_EMTPY_DATA)
    LOGGER.info(
        "handling entity loan request: {args}, {data}".format(args=args,
                                                              data=data))

    return func(data)

def get_borrowed_items_from(loan)-> db.Container:
    """Retrieves and returns the Records of items that were borrowed

    Paramters
    ---------
    loan: db.Record, borrowed items of this loan are retrieved and returned

    Returns
    -------
    db.Container, borrowed items referenced by the given loan. Return value is a container with
    one element, even if only one item is referenced and the property value is not a list.
    """
    borrowed = loan.get_property(BOX_BORROWED)
    if borrowed is None:
        raise RuntimeError(
            f"Loan has no reference to a borrowed item. {BOX_BORROWED} property is missing."
            f"Loan ID: {loan.id}")
    elif not isinstance(borrowed.value, list):
        borrowed = [borrowed.value]
    else:
        borrowed = borrowed.value
    recs = db.Container().extend([db.Record(id=rec_id.split("@")[0]) for rec_id in borrowed])
    return recs.retrieve()

def set_location_of_borrowed_items(loan, kind)-> list(db.Record):
    """Retrieve Records of borrowed items, sets the location and returns the container

    Parameters
    ----------
    kind: db.Property, should be RETURNLOCATION or DESTINATION
    """
    if loan.get_property(kind) is None:
        raise RuntimeError(
            f"Cannot set {kind} because the loan does not have one. Loan ID: {laon.id}")
    borrowed = get_borrowed_items_from(loan)
    for item in borrowed:
        set_property(item, LOCATION, loan.get_property(kind).value)
    return borrowed

def main(func, args=None):
    # configure_server_side_logging(LOGGER_NAME)
    try:
        code = _caller(func, args if args is not None else sys.argv[1:])
    except SystemExit:
        raise
    except DataModelError as e:
        print_error(e.args[0])
        code = 3
        LOGGER.debug(e)
    except DataError as e:
        print_error(e.args[0])
        code = 4
        LOGGER.debug(e)
    except BaseException as e:  # pylint: disable=W0703
        tb = traceback.format_exc()
        print_error("An unknown error occured."
                    "<br>{}<code>{}</code>".format(str(e), tb))
        code = 1
        LOGGER.debug("Unknown error!!!", e)
    LOGGER.info("exiting with code {}".format(code))
    sys.exit(code)
