#!/usr/bin/env python3
# encoding: utf-8
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#
"""
Creates a loan request using information provided by a web formular.
"""
from __future__ import absolute_import

import linkahead as db
from caosadvancedtools.serverside.helper import get_timestamp, print_success

from .box_loan import (BORROWER, BOX, COMMENT, DESTINATION, EXHAUST_CONTENTS,
                      EXPECTED_RETURN, F_BOX, F_COMMENT, F_DESTINATION,
                      F_EMAIL, F_EXHAUST_CONTENTS, F_EXPECTED_RETURN_DATE,
                      F_FIRST_NAME, F_LAST_NAME, FIRST_NAME, LAST_NAME, LOAN,
                      LOAN_REQUESTED, assert_date_in_future,
                      assert_key_in_data, get_box, insert_or_update_person, main,
                      send_loan_request_mail)


def create_loan(box, borrower, expected_return, exhaust_contents, comment,
                destination):
    """ Create a new loan record. """

    loan = db.Record().add_parent(LOAN)

    loan.add_property(BOX, box)
    loan.add_property(BORROWER, borrower)
    loan.add_property(EXPECTED_RETURN, expected_return)
    loan.add_property(EXHAUST_CONTENTS, exhaust_contents)
    loan.add_property(COMMENT, comment)
    loan.add_property(DESTINATION, destination)
    loan.add_property(LOAN_REQUESTED, get_timestamp())

    return loan


_OBLIGATORY = [
    F_EXPECTED_RETURN_DATE,
    F_BOX,
    F_COMMENT,
    F_DESTINATION,
    F_EMAIL,
    F_LAST_NAME,
    F_FIRST_NAME,
]


def _check_data(data):
    if F_EXHAUST_CONTENTS not in data:
        data[F_EXHAUST_CONTENTS] = False

    for field in _OBLIGATORY:
        assert_key_in_data(data, field)
    assert_date_in_future(
        data[F_EXPECTED_RETURN_DATE],
        ("The expected return date needs to be in the future."
         " You submitted '{}'.").format(
            data[F_EXPECTED_RETURN_DATE]))

    return data


def _issue_loan_request(data):
    """ Insert a loan record a insert/update a person record. """
    data = _check_data(data)
    borrower = insert_or_update_person(firstname=data[F_FIRST_NAME],
                          lastname=data[F_LAST_NAME],
                          email=data[F_EMAIL])
    loan = create_loan(box=data[F_BOX],
                       borrower=borrower,
                       expected_return=data[F_EXPECTED_RETURN_DATE],
                       exhaust_contents=data[F_EXHAUST_CONTENTS],
                       comment=data[F_COMMENT],
                       destination=data[F_DESTINATION])
    loan.insert()

    return borrower, loan


def issue_loan_request(data):
    """ Issue a loan request.

    I.e. inserting Record with parent `Loan` which has the following
    properties:
        * borrower
        * expectedReturn
        * exhaustContents
        * Box
        * comment
        * destination

    The borrower is a Person Record, with firstName, lastName and email,
    identified by either email, oder firstName+lastName.
    """
    borrower, loan = _issue_loan_request(data)

    fn = borrower.get_property(FIRST_NAME.name).value
    ln = borrower.get_property(LAST_NAME.name).value
    print_success('Thank you for your loan request, {fn} {ln}. See '
                  '<a href="{loan}" title="Go to the newly created loan '
                  'request.">here</a>'.format(fn=fn, ln=ln, loan=loan.id))

    send_loan_request_mail(data, borrower, loan)

    return 0


if __name__ == "__main__":
    main(issue_loan_request)
